from __future__ import print_function
from __future__ import division
from .version import __version__
from mcp3208 import MCP3208
from .Engine import Engine
from .Gait import DiscreteRippleGait
from .Leg import Leg, LegException
from .Servo import Servo

__author__ = 'Kevin J. Walchko'
__license__ = 'MIT'



#!/usr/bin/env python

##############################################
# The MIT License (MIT)
# Copyright (c) 2017 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import print_function, division
from pyxl320 import ServoSerial
from pyxl320.Packet import le, makeReadPacket
import argparse
import simplejson as json


def writeToFile(data, filename='data.json'):
	with open(filename, 'w') as outfile:
		json.dump(data, outfile)


DESCRIPTION = """
Returns the angles of servos
"""


def handleArgs():
	parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawTextHelpFormatter)
	parser.add_argument('port', help='serial port  or \'dummy\' for testing', type=str)
	parser.add_argument('-j', '--json', help='save info to a json file, you must supply a file name: --json my_file.json', type=str)

	args = vars(parser.parse_args())
	return args


def getInfo(pkt):
	ID = pkt[4]
	angle = float((pkt[10] << 8) + pkt[9])/1023
	angle *= 300.0
	return ID, angle


def getSingle(ID, ser):
	pkt = makeReadPacket(ID, 37, le(2))
	# print('made packet:', pkt)
	ID = None
	angle = None

	ans = ser.sendPkt(pkt)
	if ans:
		ans = ans[0]
		ID, angle = getInfo(ans)

	return ID, angle


def main():
	args = handleArgs()
	port = args['port']

	s = ServoSerial(port=port)
	s.open()

	ids = range(1, 13)

	resp = {}
	for k in ids:
		resp[k] = None

	# as more servos add up, I might need to increase the cnt number???
	for i in ids:
		ID, angle = getSingle(i, s)
		resp[i] = angle

	cnt = 10
	while cnt:
		cnt = 0
		for k, v in resp.items():
			# search through and find servos w/o responses (i.e., None)
			if v is None:
				cnt += 1  # found a None
				ID, angle = getSingle(k, s)
				resp[k] = angle

	print('')
	print('Servos: 1 - 12')
	print('All angles are in {}'.format('degrees'))
	print(' {:>13} | {:>13} | {:>13} | {:>13} |'.format('Leg 1', 'Leg 2', 'Leg 3', 'Leg 4'))
	print(' {:>4} | {:6} | {:>4} | {:6} | {:>4} | {:6} | {:>4} | {:6} |'.format('ID', 'Angle', 'ID', 'Angle', 'ID', 'Angle', 'ID', 'Angle'))
	print('-' * 65)
	# for k, v in resp.items():
	# 	if v is None:
	# 		print('{:4} | {}'.format(k, 'unknown'))
	# 	else:
	# 		print('{:4} | {:6.2f}'.format(k, v))
	for i in range(1, 4):
		print(' {:4} | {:6.2f} | {:4} | {:6.2f} | {:4} | {:6.2f} | {:4} | {:6.2f}'.format(i, resp[i], i+3, resp[i+3], i+6, resp[i+6], i+9, resp[i+9]))
	print('-' * 65)
	print('')

	if args['json']:
		print('Saving servo angle info to {}'.format(args['json']))
		writeToFile(resp, args['json'])

	s.close()


if __name__ == '__main__':
	main()



#!/usr/bin/env python

##############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import print_function
from __future__ import division
import numpy as np
from math import cos, sin, sqrt, pi
# from math import radians as d2r
# from Servo import Servo
# import time


def rot_z_tuple(t, c):
	"""
	t - theta [radians]
	c - [x,y,z]
	return - (x,y,z) tuple rotated about z-axis
	"""
	ans = (
		c[0]*cos(t)-c[1]*sin(t),
		c[0]*sin(t)+c[1]*cos(t),
		c[2]
	)

	return ans


# make a static method in Gait? Nothing else uses it
def rot_z(t, c):
	"""
	t - theta [radians]
	c - [x,y,z]
	return - [x,y,z] numpy array rotated about z-axis
	"""
	ans = np.array([
		c[0]*cos(t)-c[1]*sin(t),
		c[0]*sin(t)+c[1]*cos(t),
		c[2]
	])
	# ans = np.array(rot_z_tuple(t, c))

	return ans


class Gait(object):
	"""
	Base class for gaits. Gait only plan all foot locations for 1 complete cycle
	of the gait.
	"""
	# these are the offsets of each leg
	legOffset = [0, 6, 3, 9]
	# frame rotations for each leg
	cmrot = [pi/4, -pi/4, -3*pi/4, 3*pi/4]
	frame = [-pi/4, pi/4, 3*pi/4, -3*pi/4]  # this seem to work better ... wtf?
	moveFoot = None
	rest = None
	scale = 50.0

	def __init__(self, rest):
		# the resting or idle position/orientation of a leg
		self.rest = rest
		# self.moveFoot = mf
		# self.write = w

	def command(self, cmd):
		"""
		func is the quadruped move foot function for a specific leg
		"""
		x, y, rz = cmd
		d = sqrt(x**2+y**2)

		# commands should be unit length, oneCyle scales it
		if 1.0 < d:
			x /= d
			y /= d

		# handle no movement command ... do else where?
		if d < 0.1 and abs(rz) < 0.1:
			x = y = 0.0
			rz = 0.0

			return None

		# if abs(rz) < 0.001:
		# 	rz = 0.0

		return self.oneCycle(x, y, rz)

	def oneCycle(x, y, rz):
		print('*** wrong function! ***')
		return None


class DiscreteRippleGait(Gait):
	"""
	Discrete 12 step gait
	"""
	steps = 0

	def __init__(self, h, r):
		Gait.__init__(self, r)
		self.phi = [9/9, 6/9, 3/9, 0/9, 1/9, 2/9, 3/9, 4/9, 5/9, 6/9, 7/9, 8/9]  # foot pos in gait sequence
		maxl = h  # lifting higher gives me errors
		minl = maxl/2
		self.z = [minl, maxl, minl, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]  # leg height
		self.steps = len(self.phi)

	def eachLeg(self, index, cmd):
		"""
		interpolates the foot position of each leg
		cmd:
			linear (mm)
			angle (rads)
		"""
		rest = self.rest
		i = index
		phi = self.phi[i]
		xx, yy, rzz = cmd

		# rotational commands -----------------------------------------------
		angle = rzz/2-rzz*phi
		rest_rot = rot_z(-angle, rest)

		# create new move command
		move = np.array([
			xx/2 - phi*xx,
			yy/2 - phi*yy,
			self.z[i]
		])

		# new foot position: newpos = rot + move ----------------------------
		newpos = move + rest_rot
		# print('New  [](x,y,z): {:.2f}\t{:.2f}\t{:.2f}'.format(newpos[0], newpos[1], newpos[2]))
		return newpos

	def oneCycle(self, x, y, rz):
		scale = self.scale
		cmd = (scale*x, scale*y, rz)
		ret = []  # 4 leg foot positions for the entire 12 count cycle is returned

		for i in range(0, self.steps):  # iteration, there are 12 steps in gait cycle
			footPos = []
			for legNum in [0, 2, 1, 3]:  # order them diagonally
				# rcmd = self.calcRotatedOffset(cmd, legNum)
				rcmd = rot_z_tuple(self.frame[legNum], cmd)
				index = (i + self.legOffset[legNum]) % 12
				pos = self.eachLeg(index, rcmd)  # move each leg appropriately
				# print('Foot[{}]: {:.2f} {:.2f} {:.2f}'.format(legNum, *(pos)))
				# if legNum == 0: print('New  [{}](x,y,z): {:.2f}\t{:.2f}\t{:.2f}'.format(i, pos[0], pos[1], pos[2]))
				footPos.append([index, legNum, pos])  # all in leg frame
			# print('footPos', footPos)

			# corr = Correction()
			# c = corr.calcCorrection(footPos)
			# feet = corr.rotateFeetCorrected(footPos, c)

			ret.append(footPos)  # 4 feet at index i: [index, legNum, footposition]

# <<<<<<< HEAD:quadruped/Gait.py
		return ret
# =======
# 			feet = footPos
# 			# print('----------------------------')
# 			for foot in feet:
# 				legNum = foot[1]
# 				ft = foot[2]
# 				self.moveFoot(legNum, ft)
# 				# print('Foot[{}]: {:.2f} {:.2f} {:.2f}'.format(legNum, *ft))
#
# 			# Servo.bulkWrite(Servo.ser)
# 			# Servo.syncWrite(Servo.ser)
# 			self.write()
# 			time.sleep(0.1)
# >>>>>>> master:quadruped/quadruped/Gait.py


# class ContinousRippleGait(Gait):
# 	alpha = 1.0
#
# 	def __init__(self, h, r):
# 		Gait.__init__(self)
# 		self.height = h
# 		self.rest = r
#
# 	@staticmethod
# 	def phi(x):
# 		"""
# 		The phase
# 		"""
# 		phi = 0.0
# 		if x <= 3.0:
# 			phi = 1/3*(3.0-x)
# 		else:
# 			phi = 1/9*(x-3)
# 		return phi
#
# 	def z(self, x):
# 		"""
# 		Leg height
#
# 		duty cycle:
# 			0-3: leg lifted
# 			3-12: leg on ground
# 			duty = (12-3)/12 = 0.75 = 75% a walking gait
# 		"""
# 		height = self.height
# 		z = 0.0
# 		if x <= 1:
# 			z = height/1.0*x
# 		elif x <= 2.0:
# 			z = height
# 		elif x <= 3.0:
# 			z = -height/1.0*(x-2.0)+height
# 		return z
#
# 	def eachLeg(self, index, cmd):
# 		"""
# 		interpolates the foot position of each leg
# 		"""
# 		rest = self.rest
# 		i = (index*self.alpha) % 12
# 		phi = self.phi(i)
# 		z = self.z(i)
#
# 		# rotational commands -----------------------------------------------
# 		angle = cmd['angle']/2-cmd['angle']*phi
# 		rest_rot = rot_z(-angle, rest)
# 		# rest_rot[2] = 0  # let linear handle z height
#
# 		# linear commands ----------------------------------------------------
# 		linear = cmd['linear']
# 		xx = linear[0]
# 		yy = linear[1]
#
# 		# create new move command
# 		move = np.array([
# 			xx/2 - phi*xx,
# 			yy/2 - phi*yy,
# 			z
# 		])
#
# 		# new foot position: newpos = rot + move ----------------------------
# 		newpos = move + rest_rot
# 		return newpos



from __future__ import print_function, division
from pyxl320.xl320 import ErrorStatusMsg


class PacketDecoder(object):
	"""
	move to pyxl320.Packet?

	Given a packet (array of integers) from the servo, decode it.

	ToDo:
		- finish decoding all register info
	"""
	F = 0
	C = 1

	def __init__(self, pkt, offset=0):
		self.id = pkt[4]  # save servo ID
		self.instr = pkt[7]  # should be 0x55 (85) for status packet
		self.len = self.get16b(pkt[5], pkt[6]) - 4  # don't count: instr, error, crc[0,1]
		self.pkt = pkt[9:-2]  # remove header, crc, and other stuff ... keep only data section
		self.offset = offset
		self.error = pkt[8]  # should be 0 if all is well

	def checkError(self):
		"""
		Does the return status packet say there is an error?

		return:
			True: there is an error and prints info to std out
			False: all is good with the world ... turtles all the way down
		"""
		if self.instr == 0x55:
			if self.error == 0x00:
				return False
			else:
				print('Servo[{}]: {}'.format(self.id, ErrorStatusMsg[self.error]))
				return True
		else:
			print('Servo[{}]: did not give a status message'.format(self.id))
			return True
		return True

	@staticmethod
	def get16b(lo, hi):
		return ((hi << 8) + lo)

	def getBase(self, reg):
		base = reg - self.offset
		# print('base', base)

		if base < 0:
			raise Exception('PacketDecoder::base < 0')

		if base >= self.len:
			raise Exception('PacketDecoder::packet not long enough to access this packet')

		return base

	def voltage(self):
		base = self.getBase(45)
		return self.pkt[base]/10

	def angle(self):
		base = self.getBase(37)
		return self.get16b(self.pkt[base], self.pkt[base+1])/1023 * 300.0

	def load(self):
		base = self.getBase(41)
		load = self.get16b(self.pkt[base], self.pkt[base+1])
		direction = 'CW' if (load & 1024) == 1024 else 'CCW'
		percent = (load & 1023)/1023 * 100
		return percent, direction

	def temperature(self, scale=0):
		base = self.getBase(46)
		temp = 0.0
		if scale == self.F:
			temp = self.pkt[base]*9/5+32
		else:
			temp = self.pkt[base]*1.0
		return temp

	def hw_error(self):
		base = self.getBase(50)
		return self.pkt[base]

	def printPacket(self):
		"""
		"""
		print('-'*70)
		print('Servo[{}]  instr: {}  error: {}  len(data): {}'.format(self.id, self.instr, self.error, self.len))
		print('raw:', self.pkt)




from __future__ import print_function
from __future__ import division
import numpy as np
from numpy.linalg import norm
# import logging
from math import cos
# from math import sin
# from math import sqrt
from math import pi
from Gait import rot_z


class RobotTransform(object):
	"""
	Converts pts between the leg and body frame.
	"""
	def __init__(self, radius):
		"""
		in:
			radius - radius of body from CM in mm
			anglesB2L - angles to transform between body and leg frames
		"""
		# cmrot
		self.leg2body = [pi/4, -pi/4, -3*pi/4, 3*pi/4]  # legs to body frame # orig
#         self.leg2body = [-pi/4, pi/4, 3*pi/4, -3*pi/4]  # legs to body frame
		# frame
		self.body2leg = [-pi/4, pi/4, 3*pi/4, -3*pi/4]  # body to leg frame # orig
#         self.body2leg = [pi/4, -pi/4, -3*pi/4, 3*pi/4]  # body to leg frame
		# account for base, in base frame
		cm = radius*cos(pi/4)
		self.base = [
			np.array([cm, cm, 0]),
			np.array([cm, -cm, 0]),
			np.array([-cm, -cm, 0]),
			np.array([-cm, cm, 0])
		]

	def leg2Body(self, legNum, pts):
		"""
		Converts points from leg_frame to body_frame
		"""
		pts2 = rot_z(self.leg2body[legNum], pts) + self.base[legNum]
		return pts2

	def body2Leg(self, legNum, pts):
		"""
		Converts points from body_frame to leg_frame
		"""
		pts2 = rot_z(self.body2leg[legNum], pts-self.base[legNum])
		return pts2


class Correction(object):
	def __init__(self):
		pass

	@staticmethod
	def inside(feet, prnt=False):
		"""
		Determine if a point P is inside of a triangle composed of points
		A, B, and C.
		pts = [A,B,C]
		P = [0,0] at the center of mass of the robot
		returns True (inside triangle) or False (outside the triangle)
		"""
		pts = []
		for p in feet:
			if isinstance(p, np.ndarray):
				pts.append(p)

		# print('inSideCM pts:', pts)
		A = pts[0][0:2]
		B = pts[1][0:2]
		C = pts[2][0:2]
		P = np.array([0, 0])  # CM is at the center :)

		# Compute vectors
		v0 = C - A
		v1 = B - A
		v2 = P - A

		# Compute dot products
		dot00 = np.dot(v0, v0)
		dot01 = np.dot(v0, v1)
		dot02 = np.dot(v0, v2)
		dot11 = np.dot(v1, v1)
		dot12 = np.dot(v1, v2)

		# Compute barycentric coordinates
		invDenom = 1 / (dot00 * dot11 - dot01 * dot01)
		u = (dot11 * dot02 - dot01 * dot12) * invDenom
		v = (dot00 * dot12 - dot01 * dot02) * invDenom

		if prnt:
			print(u, v)

		# Check if point is in triangle
		ans = ((u >= 0) and (v >= 0) and (u + v < 1))

		print('inside():', ans)
		return ans

	@staticmethod
	def lineIntersection(p1, p2, p3):
		"""
		Find the intersection of 2 lines.
		line 1: p1, p2
		line 2: p3, [0,0]
		"""
		x1 = p1[0]
		x2 = p2[0]
		x3 = p3[0]
		x4 = 0.0
		y1 = p1[1]
		y2 = p2[1]
		y3 = p3[1]
		y4 = 0.0

		denom = ((x1 - x2)*(y3 - y4) - (y1 - y2)*(x3 - x4))
		if abs(denom) < 0.00001:
			# print('crap {}'.format(denom))
			return np.array([0, 0])
		x = ((x1*y2 - y1*x2)*(x3 - x4) - (x1 - x2)*(x3*y4 - y3*x4))/denom
		y = ((x1*y2 - y1*x2)*(y3 - y4) - (y1 - y2)*(x3*y4 - y3*x4))/denom
		return np.array([x, y])

	@staticmethod
	def vmin(a):
		"""
		Find the minimum vector in an array of 2D vectors.
		in = [[1,2], [2,3], ...]
		out = [1,2]
		"""
		minv = 0
		min_val = 1000000000000000
		for p in a:
			val = norm(p)
			if val < min_val:
				min_val = val
				minv = p
		return minv

	def correction(self, feet, movingFoot):
		"""
		Given the robot's foot locations, provide correction if the
		center of mass (CM) is outside the triangle formed by the 3
		foot locations.
		pts = [foot0, foot1, foot2, foot3]
		correction = [x,y,0]
		"""
#         a = []
#         for i in range(0, 3):
#             p0 = pts[i]
#             p1 = pts[(i+1) % 3]
#             xx = self.lineIntersection(p0, p1, cm0, cm1)
#             a.append(xx)
#         a = self.vmin(a)
#         correction = np.array([-a[0], -a[1], 0.0])
		# 0 1 2 3
		op = feet[(movingFoot + 2) % 4]
		p0 = feet[(movingFoot + 1) % 4]
		p1 = feet[(movingFoot + 3) % 4]
		a = self.lineIntersection(p0, p1, op)
		correction = np.array([-a[0], -a[1], 0.0])
		return correction

	def calcCorrection(self, feet):
		"""
		This take the feet positions, calculated by a Gait, and adjusts them
		to ensure the CM is inside of the stability triangle.

		in:
			feet = [[index, legNum, foot], ...] there are 4 of these
					index - tells if foot is up/down
					legNum - which leg [0-3]
					foot - (x,y,z) is converted to 2D
		out:
			correction = [x,y,0] is only a 2D correction
		"""
		temp = [0, 0, 0, 0]
		tf = RobotTransform(45)

		print('-------------------')
		print('calcCorrection()')
		movingFoot = 0
		for f in feet:
			index = f[0]
			foot = f[2]
			legNum = f[1]
			if index > 2:
				print('in foot: {:.2f} {:.2f}'.format(*foot[0:2]))
				ft = tf.leg2Body(legNum, foot)
				temp[legNum] = ft
				print('rotated foot: {:.2f} {:.2f}'.format(*ft[0:2]))
			else:
				movingFoot = legNum

		if not self.inside(temp):
			correction = 1.5*self.correction(temp, movingFoot)
		else:
			correction = np.array([0, 0, 0])
		return correction

	def rotateFeetCorrected(self, feet, correction):
		"""
		return: [[index, legNum, (x,y,z)], ...] corrected stationary feet positions
		"""
		tf = RobotTransform(45)
		ans = []
		for p in feet:  # p = [index, legNum, (x,y,z)]
			index = p[0]
			legNum = p[1]
			foot = p[2]
			if index > 2:  # check index to see if leg moving
				foot = tf.leg2Body(legNum, foot) + correction
				foot = tf.body2Leg(legNum, foot)
			ans.append([index, legNum, foot])
		return ans



#!/usr/bin/env python
##############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import print_function
from __future__ import division
import numpy as np
from math import sin, cos, acos, atan2, sqrt, pi, fabs
from math import radians as d2r
from math import degrees as r2d
# import logging
from .Servo import Servo

# logging.basicConfig(level=logging.DEBUG)
# logging.basicConfig(level=logging.ERROR)


class LegException(Exception):
	pass


class Leg(object):
	"""
	"""
	# these are fixed by the 3D printing, not changing
	coxaLength = 45.0
	tibiaLength = 55.0
	femurLength = 104.0
	s_limits = [
		[-90, 90],  # set limits
		[-90, 90],
		[-150, 0]
	]
	s_offsets = [150, 150, 150+90]  # angle offsets to line up with fk

	sit_raw = (150, 270, 100)
	stand_raw = (150, 175, 172)
	sit_angles = None
	stand_angles = None

	def __init__(self, channels):
		"""
		Each leg has 3 servos/channels
		"""
		if not len(channels) == 3:
			raise LegException('len(channels) != 3')

		Servo.bulkServoWrite = True

		# angle offsets to line up with fk
		self.servos = []
		for i in range(0, 3):
			self.servos.append(Servo(channels[i]))
			self.servos[i].setServoLimits(self.s_offsets[i], *self.s_limits[i])

		self.sit_angles = self.convertRawAngles(*self.sit_raw)
		# initAngles = [0, 0, -90+30]  # nico legs have a small offset
		# initAngles = [0, 45, -90+30-45]  # nico legs have a small offset
		initAngles = self.convertRawAngles(*self.stand_raw)
		self.stand_angles = initAngles
		self.foot0 = self.fk(*initAngles)  # rest/idle position of the foot/leg
		# print('foot0', self.foot0)

	def __del__(self):
		pass

	def sit(self):
		self.moveFootAngles(*self.sit_angles)

	def stand(self):
		self.moveFootAngles(*self.stand_angles)

	def convertRawAngles(self, a, b, c):
		return (a-self.s_offsets[0], b-self.s_offsets[1], c-self.s_offsets[2])

	def fk(self, a, b, g):
		"""
		Forward kinematics of the leg, note, angle are all degrees
		"""
		Lc = self.coxaLength
		Lf = self.femurLength
		Lt = self.tibiaLength

		a = d2r(a)
		b = d2r(b)
		g = d2r(g)

		foot = [
			(Lc + Lf*cos(b) + Lt*cos(b + g))*cos(a),
			(Lc + Lf*cos(b) + Lt*cos(b + g))*sin(a),
			Lf*sin(b) + Lt*sin(b + g)
		]

		return np.array(foot)

	def ik(self, x, y, z):
		"""
		Calculates the inverse kinematics from a given foot coordinate (x,y,z)[mm]
		and returns the joint angles[degrees]

		Reference (there are typos)
		https://tote.readthedocs.io/en/latest/ik.html

		If the foot (x,y,z) is in a position that does not produce a result (some
		numeric error or invalid foot location), the this returns None.
		"""
		# try:
		Lc = self.coxaLength
		Lf = self.femurLength
		Lt = self.tibiaLength

		if sqrt(x**2 + y**2) < Lc:
			print('too short')
			return None
		# elif z > 0.0:
		# 	return None

		a = atan2(y, x)
		f = sqrt(x**2 + y**2) - Lc
		# b1 = atan2(z, f)  # takes into account quadrent, z is neg

		# you have different conditions depending if z is pos or neg
		if z < 0.0:
			b1 = atan2(f, fabs(z))
		else:
			b1 = atan2(z, f)

		d = sqrt(f**2 + z**2)  # <---

		# print('ik pos: {} {} {}'.format(x,y,z))
		# print('d: {:.3f}  f: {:.3f}'.format(d,f))
		# print('Lc Lf Lt: {} {} {}'.format(Lc,Lf,Lt))
		# print('num: {:.3f}'.format(Lf**2 + d**2 - Lt**2))
		# print('den: {:.3f}'.format(2.0 * Lf * d))
		# print('acos: {:.2f}'.format((Lf**2 + d**2 - Lt**2) / (2.0 * Lf * d)))

		guts = ((Lf**2 + d**2 - Lt**2) / (2.0 * Lf * d))
		if 1.0 < guts or guts < -1.0:
			print('acos crap!: {:.3f} {:.3f} {:.3f} guts: {:.3f}'.format(x, y, z, guts))
			return None
		b2 = acos((Lf**2 + d**2 - Lt**2) / (2.0 * Lf * d))  # issues?
		b = b1 + b2
		g = acos((Lf**2 + Lt**2 - d**2) / (2.0 * Lf * Lt))

		# FIXES ###################################
		g -= pi  # fix to align fk and ik frames

		if z < 0.0:
			b -= pi/2  #
		##############################################
		# print('b1 b2: {:.2f} {:.2f}'.format(r2d(b1), r2d(b2)))
		# print('ik angles: {:.2f} {:.2f} {:.2f}'.format(r2d(a), r2d(b), r2d(g)))
		return r2d(a), r2d(b), r2d(g)  # coxaAngle, femurAngle, tibiaAngle

		# except Exception as e:
		# 	print('ik error:', e)
		# 	raise e

	def moveFoot(self, x, y, z):
		"""
		Attempts to move it's foot to coordinates [x,y,z]
		"""
		try:
			# a, b, c = self.ik(x, y, z)  # inverse kinematics
			# angles = [a, b, c]
			angles = self.ik(x, y, z)  # inverse kinematics
			# return angles
			if angles is None:
				print('something bad')
				return
			# print('angles: {:.2f} {:.2f} {:.2f}'.format(*angles))
			for i, servo in enumerate(self.servos):
				# print('i, servo:', i, servo)
				angle = angles[i]
				# print('servo {} angle {}'.format(i, angle))
				servo.angle = angle
			return angles

		except Exception as e:
			print (e)
			raise

	def moveFootAngles(self, a, b, c):
		"""
		Attempts to move it's foot to coordinates [x,y,z]
		"""
		try:
			for servo, angle in zip(self.servos, (a, b, c)):
				servo.angle = angle

		except:
			print('Leg::moveFootAngles() error')
			raise

	# def reset(self):
	# 	# self.angles = self.resting_position
	# 	self.move(*self.foot0)


if __name__ == "__main__":
	print('Hello cowboy!')



#!/usr/bin/env python

##############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import print_function
from __future__ import division
from .Leg import Leg
from pyxl320 import ServoSerial
# from pyxl320 import DummySerial
from .Servo import Servo
from pyxl320 import Packet
from pyxl320 import xl320
import time


class Engine(object):
	"""
	change name to Hardware???

	This is the low level driver that can be executed w/o using pyGecko.
	"""
	def __init__(self, data={}):
		"""
		Sets up all 4 legs and servos. Also setups limits for angles and servo
		pulses.
		"""
		if 'serialPort' in data:
			try:
				ser = ServoSerial(data['serialPort'])
				print('Using servo serial port: {}'.format(data['serialPort']))

			except:
				print('bye ...')
				exit(1)
		else:
			print('Using dummy serial port!!!')
			ser = ServoSerial('test_port', fake=True)
			# raise Exception('No serial port given')

		ser.open()
		Servo.ser = ser  # set static serial port, not sure I like this

		if 'write' in data:
			method = data['write']
			if method == 'sync':
				Servo.syncServoWrite = True  # FIXME: this is broken
				print('*** using sync write ***')
			elif method == 'bulk':
				Servo.bulkServoWrite = True
				print('*** using bulk write ***')

		self.legs = []
		for i in range(0, 4):  # 4 legs
			channel = i*3  # 3 servos per leg
			self.legs.append(
				Leg([channel+1, channel+2, channel+3])
			)

		# better way?????
		self.servoWrite = self.legs[0].servos[0].write

		self.stand()

	def __del__(self):
		"""
		Leg kills (reboots) all servos on exit.

		Eventually will put the robot in sit pose.
		"""
		self.sit()
		pkt = Packet.makeRebootPacket(xl320.XL320_BROADCAST_ADDR)
		Servo.ser.write(pkt)
		Servo.ser.write(pkt)
		time.sleep(0.1)
		Servo.ser.close()  # close static serial port

	def sit(self):
		"""
		sequence to sit down nicely
		"""
		for leg in self.legs:
			leg.sit()
		self.servoWrite()  # FIXME: ugly
		time.sleep(1)

	def stand(self):
		"""
		sequence to stand up nicely
		"""
		for leg in self.legs:
			leg.stand()
		self.servoWrite()  # FIXME: ugly
		time.sleep(1)

	def getFoot0(self, i):
		"""
		rename getNeutral?
		"""
		return self.legs[i].foot0

	def move(self, cycle):
		"""
		cycle - move of 4 legs through all steps
		"""
		for mov in cycle:
			for leg in mov:
				index = leg[1]
				footPos = leg[2]
				# print('Leg[{}]: {}'.format(index, footPos))
				self.legs[index].moveFoot(*footPos)
			self.servoWrite()  # FIXME: ugly
			time.sleep(0.1)



#!/usr/bin/env python
##############################################
# The MIT License (MIT)
# Copyright (c) 2016 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import print_function
from __future__ import division
import logging
from pyxl320 import Packet
from pyxl320.Packet import makeSyncAnglePacket
# from pyxl320 import DummySerial
from pyxl320 import xl320

logger = logging.getLogger(__name__)


def makeBulkAnglePacket(info):
	"""
	Write bulk angle information to servos.

	info = [[ID, angle], [ID, angle], ...]
	"""
	addr = Packet.le(xl320.XL320_GOAL_POSITION)
	data = []
	for pkt in info:
		data.append(pkt[0])  # ID
		data.append(addr[0])  # LSB
		data.append(addr[1])  # MSB
		data.append(2)
		data.append(0)
		angle = Packet.le(int(pkt[1]/300*1023))
		data.append(angle[0])  # LSB
		data.append(angle[1])  # MSB

	ID = xl320.XL320_BROADCAST_ADDR
	instr = xl320.XL320_BULK_WRITE
	pkt = Packet.makePacket(ID, instr, None, data)  # create packet

	# print(pkt)
	return pkt


# I don't like global variables, but I don't know how else to have one array
# shared by all servos and have one simple class method to push updates
gBulkData = []
gSyncData = []  # this is much less data than bulk


class ServoBase(object):
	"""
	I might reverse these classes, so i can combine RC and Robotis servos
	ServoBase
	|  angle()
	|  limits()
	+- RCServo
		|  servo things
	+- RobotisServo
		|  bulkWrite
	"""
	ser = None
	bulkServoWrite = False
	syncServoWrite = False
	# bulkData = [0]

	def __init__(self):
		# just put a dummy serial here, because testing needs it in the angle setter
		if self.ser is None:
			# self.ser = DummySerial('test')
			raise Exception('ServoBase::__init__() no serial port')

	def bulkWrite(self):
		if self.ser is None:
			raise Exception('bulkWrite no serial port')
		# print('Servo::bulkWrite()')
		global gBulkData
		pkt = makeBulkAnglePacket(gBulkData)
		self.ser.write(pkt)
		gBulkData = []  # clear global

	def syncWrite(self):
		if self.ser is None:
			raise Exception('syncWrite no serial port')
		# print('Servo::syncWrite()')
		global gSyncData
		# print('gSyncData', gSyncData)
		pkt = makeSyncAnglePacket(gSyncData)
		# print('pkt', pkt)
		self.ser.write(pkt)
		# self.ser.write(pkt)
		gSyncData = []

	def write(self):
		if self.bulkServoWrite:
			self.bulkWrite()
		elif self.syncServoWrite(self.ser):
			self.syncWrite(self.ser)
		else:
			raise Exception('ServoBase::write() not bulk or sync')


class Servo(ServoBase):
	"""
	Keeps info for servo and commands their movement.
	angles are in degrees

				0
	-150 ------ + ------- 150  kinematics (servo_k)
				150
		0 ------ + ------- 300  real servo (servo_r)

	servo_k commands are between -150 and 150 degrees, with 0 deg being center

	servo to kinematics: servo_r = servo_k + 150
	kinematics to servo: servo_k = servo_r - 150
	"""
	_angle = 0.0  # current angle
	_offset = 150.0  # angle offset default, see above for explaination
	minAngle = -150.0
	maxAngle = 150.0
	ID = 0

	def __init__(self, ID):
		"""
		limits [angle, angle] - [optional] set the angular limits of the servo to avoid collision
		"""
		ServoBase.__init__(self)
		self.ID = ID
		self.setServoLimits(150.0, -150.0, 150.0)  # defaults: offset, min, max

	@property
	def angle(self):
		"""
		Returns the current servo angle
		"""
		return self._angle

	@angle.setter
	def angle(self, angle):
		"""
		Sets the servo angle and clamps it between [limitMinAngle, limitMaxAngle].
		It also commands the servo to move.
		"""

		# saturates the angle if it is outside of the limits
		if self.minAngle > angle > self.maxAngle:
			# raise Exception('@angle.setter {} > {} > {}'.format(self.minAngle, angle, self.maxAngle))
			print('@angle.setter error {} > {} > {}'.format(self.minAngle, angle, self.maxAngle))

			if self.minAngle > angle:
				angle = self.minAngle
			elif self.maxAngle < angle:
				angle = self.maxAngle

			print('@angle.setter now {}'.format(angle))

		# print('servo[{}]: bulkWrite {}'.format(self.ID, self.bulkServoWrite))

		# only send a pkt if there is a change
		if self._angle != angle:
			self._angle = angle
			# servos are centered at 150 deg, but offset can change
			# print(self.ID, angle, angle + self._offset)

			if self.syncServoWrite:
				# global gBulkData
				# gBulkData.append([self.ID, angle + self._offset])
				gSyncData.append([self.ID, angle + self._offset])
				# self.bulkData.append([self.ID, angle + self._offset])
				# print('servo[{}]: bulkWrite {}'.format(self.ID, gBulkData))
			elif self.bulkServoWrite:
				# global gBulkData
				gBulkData.append([self.ID, angle + self._offset])
				# gSyncData.append([self.ID, angle + self._offset])
				# self.bulkData.append([self.ID, angle + self._offset])
				# print('servo[{}]: bulkWrite {}'.format(self.ID, gBulkData))
			else:
				pkt = Packet.makeServoPacket(self.ID, angle + self._offset)
				self.ser.sendPkt(pkt)
				# print('sent angle {} [{}]'.format(angle, angle + self._offset))

	def setServoLimits(self, offset, minAngle, maxAngle):
		"""
		sets maximum and minimum achievable angles. Remeber, the limits have to
		be within the servo range of [0, 180] ... anything more of less won't
		work unless your change setServoRangleAngle() to something other than
		0 - 180.

		in:
			minAngle - degrees
			maxAngle - degrees
		"""
		if minAngle > maxAngle:
			raise Exception("Servo::setServoLimits: min angle > max angle")

		self._offset = offset
		self.minAngle = minAngle
		self.maxAngle = maxAngle

		# # pktmax, pktmin = Packet.makeServoLimits(self.ID, maxAngle, minAngle)
		# angle = int(maxAngle/300.0*1023)
		# pkt = Packet.makeWritePacket(self.ID, xl320.XL320_CCW_ANGLE_LIMIT, Packet.le(angle))
		# ser.write(pkt)
		# # # ser.read()
		# angle = int(minAngle/300.0*1023)
		# pkt = Packet.makeWritePacket(self.ID, xl320.XL320_CW_ANGLE_LIMIT, Packet.le(angle))
		# ser.write(pkt)
		# # ser.read()

		# if 0:
		# 	# Example: kinematics only allow a servo between -180 and 0
		# 	# min -180
		# 	# max 0
		# 	# offset 240  <-- a real servo angle of 240 is a kinematic angle of 0
		# 	# limits = [-180+240, 0+240] = [60, 240]  <-- these are real servo angles
		# 	pkt = Packet.makeServoMinLimitPacket(self.ID, minAngle+self._offset)
		# 	self.ser.sendPkt(pkt)
		# 	pkt = Packet.makeServoMaxLimitPacket(self.ID, maxAngle+self._offset)
		# 	self.ser.sendPkt(pkt)


if __name__ == "__main__":
	print('Hello space cowboy!')



#!/usr/bin/env python

##############################################
# The MIT License (MIT)
# Copyright (c) 2017 Kevin Walchko
# see LICENSE for full details
##############################################

from __future__ import print_function, division
from pyxl320 import ServoSerial
from pyxl320.Packet import le, makeReadPacket
import argparse
import simplejson as json
from serial import SerialException
import sys
from packetDecoder import PacketDecoder


def writeToFile(data, filename='data.json'):
	with open(filename, 'w') as outfile:
		json.dump(data, outfile)


def pktToDict(p):
	# print('pkt', pkt)
	# print('len(pkt)', len(pkt))
	ans = {}
	ans['ID'] = p.id
	ans['Present Position'] = p.angle()
	ans['Present Voltage'] = p.voltage()
	ans['Present Load'] = '{:>5.1f}% {}'.format(*p.load())
	ans['Present Temperature'] = p.temperature(PacketDecoder.F)
	ans['Hardware Error Status'] = p.hw_error()

	return ans


DESCRIPTION = """
Returns limited info for each leg servo.

./get_leg_info.py /dev/tty.usbserial-AL034G2K
Opened /dev/tty.usbserial-AL034G2K @ 1000000

Servos: 1 - 12
--------------------------------------------------
Servo: 1  		HW Error: 0
Position [deg]: 139.6  Load:   0.0% CCW
Voltage [V]  7.0     Temperature [F]:  80.6
--------------------------------------------------
Servo: 2  		HW Error: 0
Position [deg]: 178.9  Load:   4.5% CW
Voltage [V]  7.1     Temperature [F]:  86.0
--------------------------------------------------
Servo: 3  		HW Error: 0
Position [deg]: 119.1  Load:   0.0% CCW
Voltage [V]  7.1     Temperature [F]:  80.6
--------------------------------------------------
Servo: 4  		HW Error: 0
Position [deg]: 146.6  Load:   0.8% CCW
Voltage [V]  7.3     Temperature [F]:  80.6
--------------------------------------------------
Servo: 5  		HW Error: 0
Position [deg]: 275.4  Load:   0.8% CCW
Voltage [V]  7.1     Temperature [F]:  80.6
--------------------------------------------------
Servo: 6  		HW Error: 0
Position [deg]: 104.1  Load:   0.0% CCW
Voltage [V]  7.3     Temperature [F]:  82.4
--------------------------------------------------
Servo: 7  		HW Error: 0
Position [deg]: 163.9  Load:   0.0% CCW
Voltage [V]  7.2     Temperature [F]:  80.6
--------------------------------------------------
Servo: 8  		HW Error: 0
Position [deg]: 279.5  Load:   0.0% CCW
Voltage [V]  7.1     Temperature [F]:  80.6
--------------------------------------------------
Servo: 9  		HW Error: 0
Position [deg]: 100.3  Load:   0.0% CCW
Voltage [V]  7.1     Temperature [F]:  84.2
--------------------------------------------------
Servo: 10  		HW Error: 0
Position [deg]: 156.3  Load:   0.0% CCW
Voltage [V]  7.1     Temperature [F]:  82.4
--------------------------------------------------
Servo: 11  		HW Error: 0
Position [deg]: 280.6  Load:   0.0% CCW
Voltage [V]  7.2     Temperature [F]:  80.6
--------------------------------------------------
Servo: 12  		HW Error: 0
Position [deg]:  97.7  Load:   0.0% CCW
Voltage [V]  7.1     Temperature [F]:  84.2
--------------------------------------------------
"""


def handleArgs():
	parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawTextHelpFormatter)
	parser.add_argument('port', help='serial port  or \'dummy\' for testing', type=str)
	parser.add_argument('-j', '--json', metavar='FILENAME', help='save info to a json file: --json my_file.json', type=str)

	args = vars(parser.parse_args())
	return args


def getSingle(ID, ser):
	pkt = makeReadPacket(ID, 37, le(50-37+1))
	# print('made packet:', pkt)

	ans = ser.sendPkt(pkt)
	if ans:
		ans = ans[0]
		pd = PacketDecoder(ans, 37)  # data packet starts at register 37
		# pd.printPacket()
		if pd.checkError():
			raise Exception('Crap!')
		ans = pktToDict(pd)
	else:
		ans = None

	return ans


def printServo(s):
	print('-'*50)
	print('Servo: {}  \t\tHW Error: {}'.format(s['ID'], s['Hardware Error Status']))
	print('Position [deg]: {:5.1f}  Load: {}'.format(s['Present Position'], s['Present Load']))
	print('Voltage [V] {:4.1f}     Temperature [F]: {:5.1f}'.format(s['Present Voltage'], s['Present Temperature']))


def main():
	args = handleArgs()
	port = args['port']

	s = ServoSerial(port=port)

	# open serial port
	try:
		s.open()
	except SerialException as e:
		print('-'*20)
		print(sys.argv[0], 'encountered an error')
		print(e)
		exit(1)

	ids = range(1, 13)

	resp = {}
	for k in ids:
		resp[k] = None

	# get servo data
	try:
		for i in ids:
			data = getSingle(i, s)
			resp[i] = data
	except Exception as e:
		print(e)
		exit(1)

	cnt = 10
	while cnt:
		cnt = 0
		for k, v in resp.items():
			# search through and find servos w/o responses (i.e., None)
			if v is None:
				cnt += 1  # found a None
				ans = getSingle(k, s)
				resp[k] = ans

	print('')
	print('Servos: 1 - 12')
	for i in range(1, 13):
		printServo(resp[i])
	print('-' * 50)
	print('')

	if args['json']:
		print('Saving servo angle info to {}'.format(args['json']))
		writeToFile(resp, args['json'])

	s.close()


if __name__ == '__main__':
	main()



#!/usr/bin/env python

from __future__ import division
from __future__ import print_function
import simplejson as json


class FileStorageError(Exception):
	pass


class jsonFile(object):
	"""
	Store your API keys or other params in a json/yaml file and then import them
	with this class. You can also pass a dictionary to this and create a json/yaml
	file storage. All key/value pairs are stored in a dictionary called db.
	"""
	db = None

	def read(self, fname):
		"""
		Reads a Json file
		in: file name
		out: length of file, dictionary
		"""
		try:
			with open(fname, 'r') as f:
				data = json.load(f)

			self.db = data
			return len(self.db), data
		except IOError:
			raise FileStorageError('Could not open {0!s} for reading'.format((fname)))

	def write(self, fname, data=None):
		"""
		Writes a Json file
		"""
		try:
			if data is None:
				data = self.db

			with open(fname, 'w') as f:
				json.dump(data, f)

		except IOError:
			raise FileStorageError('Could not open {0!s} for writing'.format((fname)))

	def __getitem__(self, keyName):
		if keyName in self.db:
			return self.db[keyName]
		else:
			return None

	def __str__(self):
		s = []
		for k, v in db.items():
			s.append(str(k) + ': ' + str(v) + '\n')
		return ''.join(s)

	def __repr__(self):
		print(self.__str__())

	def clear(self):
		self.db = None


if __name__ == '__main__':
	fs = jsonFile()
	fs.read('walkingeye.json')
	db = fs.db

	print('serialPort', fs['serialPort'])
	print('')

	print(fs)
